<h3>Contents</h3>

<ul>
  <li><a href="{{ pathto('index') }}">Home</a></li>

  <li><a href="{{ pathto('getting-started') }}">Get started</a></li>
  <li><a href="{{ pathto('how-to/index') }}">How-to guides</a></li>
  <li><a href="{{ pathto('reference/index') }}">Reference guides</a></li>
  <li><a href="{{ pathto('explanation/index') }}">Explanation</a></li>
  <li><a href="{{ pathto('contents') }}">Complete table of contents</a></li>
  <li><a href="{{ pathto('example/index') }}">Library of examples</a></li>
</ul>

<h3>About the project</h3>

<ul>
  <li><a href="{{ pathto('changelog') }}">Changelog</a></li>
  <li><a href="{{ pathto('contributing') }}">Contributing</a></li>
  <li><a href="{{ pathto('backwards-compatibility') }}">Backwards Compatibility</a></li>
  <li><a href="{{ pathto('py27-py34-deprecation') }}">Python 2.7 and 3.4 Support</a></li>
  <li><a href="{{ pathto('sponsor') }}">Sponsor</a></li>
  <li><a href="{{ pathto('tidelift') }}">pytest for Enterprise</a></li>
  <li><a href="{{ pathto('license') }}">License</a></li>
  <li><a href="{{ pathto('contact') }}">Contact Channels</a></li>
</ul>

{%- if display_toc %}
  <hr>
  {{ toc }}
{%- endif %}

<hr>
<a href="{{ pathto('genindex') }}">Index</a>
<hr>
