# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

"""
mozgeckoprofiler has utilities to symbolicate and load gecko profiles.
"""
from .profiling import save_gecko_profile, symbolicate_profile_json
from .symbolication import ProfileSymbolicator
from .viewgeckoprofile import view_gecko_profile

__all__ = [
    "save_gecko_profile",
    "symbolicate_profile_json",
    "ProfileSymbolicator",
    "view_gecko_profile",
]
